# Generated by Django 6.0.6 on 2026-07-09 19:32

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='HueyEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ts', models.FloatField(db_index=True)),
                ('queue', models.CharField(db_index=True, max_length=255)),
                ('task_id', models.CharField(max_length=255)),
                ('task', models.CharField(db_index=True, max_length=255)),
                ('signal', models.CharField(max_length=64)),
                ('duration', models.FloatField(null=True)),
                ('error', models.TextField(null=True)),
                ('args', models.TextField(null=True)),
            ],
            options={
                'verbose_name': 'event',
                'db_table': 'huey_event',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='HueyDashboard',
            fields=[
            ],
            options={
                'verbose_name': 'dashboard',
                'verbose_name_plural': 'dashboard',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('hueystats.hueyevent',),
        ),
    ]
